library line_intersection;

import "dart:collection";
import "simple_features.dart" show DirectPosition2D;
export "simple_features.dart" show DirectPosition2D;

_require(cond, [msg]) {
    if (!cond) throw new ArgumentError(msg == null ? "" : msg);
}

int orientation(p, q, r) {
  var v = (p.x - r.x) * (q.y - r.y) - (q.x - r.x) * (p.y - r.y);
  if (v < 0) return -1;
  if (v == 0) return 0;
  return 1;
}

/**
 * Represents closed line segment given by two 2D points.
 */
class LineSegment {
  DirectPosition2D _start;
  DirectPosition2D _end;

  LineSegment(start, end) {
    _require(start != end, "line segments must not be collapsed: $start, $end");
    if (start <= end) {
      _start = start;
      _end = end;
    } else {
      _start = end;
      _end = start;
    }
  }
  factory LineSegment.from(start, end) {
    _require(start is List);
    _require(end is List);
    _require(start.length == 2);
    _require(end.length == 2);
    return new LineSegment(
        new DirectPosition2D(start[0], start[1]),
        new DirectPosition2D(end[0], end[1])
    );
  }

  bool isEndPoint(DirectPosition2D p) => p == _start || p == _end;

  bool _fastExcludeIntersect(LineSegment other) {
    var minx = _start.x;
    var miny = _start.y < _end.y ? _start.y : _end.y;
    var maxx = _end.x;
    var maxy = _start.y >= _end.y ? _start.y : _end.y;
    if (other._start.x < minx && other._end.x < minx) return true;
    if (other._start.x > maxx && other._end.x > maxx) return true;
    if (other._start.y < miny && other._end.y < miny) return true;
    if (other._start.y > maxy && other._end.y > maxy) return true;
    return false;
  }

  /**
   * This [LineSegments] intersects with an [other] iff the intersection
   * consists of exactly one point.
   */
  bool intersects(LineSegment other) {
    if (this == other) return false;
    if (_fastExcludeIntersect(other)) return false;
    if (this.connectsTo(other) && !this.overlaps(other)) return true;
    var o1 = orientation(_start, _end, other._start);
    var o2 = orientation(_start, _end, other._end);
    if (o1 == o2) return false;
    o1 = orientation(other._start, other._end, _start);
    o2 = orientation(other._start, other._end, _end);
    return o1 != o2;
  }

  /**
   * This [LineSegment] conntects to [other] iff the intersection consists
   * of exactly one end point.
   */
  bool connectsTo(LineSegment other) {
    return this != other
        && (other.isEndPoint(_start) || other.isEndPoint(_end));
  }

  /**
   * This [LineSegment] is colinear with [other] iff this and [other]
   * lie on one straight line.
   *
   */
  bool isColinear(LineSegment other) {
    var o1 = orientation(_start, _end, other._start);
    var o2 = orientation(_start, _end, other._end);
    return o1 == 0 && o2 == 0;
  }

  /**
   * This [LineSegment] and [other] overlap iff their intersection
   * consists of more than one point.
   */
  bool overlaps(LineSegment other) {
    if (this == other) return true;
    if (!isColinear(other)) return false;
    if (other.start < start && start < other.end) return true;
    if (other.start < end && end < other.end) return true;
    return false;
  }

  bool operator==(LineSegment other) =>
      start == other.start && end == other.end;

  bool get isVertical => _start.x == _end.x;

  DirectPosition2D intersectionWithSweepline(num x) {
    if (isVertical) throw new StateError("intersection with vertical line not possible");
    if (x < _start.x || x > _end.x) return null;
    if (x == _start.x) return _start;
    if (x == _end.x) return _end;
    var dy = slope * (x - _start.x);
    return new DirectPosition2D(x,_start.y + dy);
  }

  DirectPosition2D intersectionWith(LineSegment other) {

  }

  double get slope => (_end.y - _start.y).toDouble() / (_end.x - _start.x).toDouble();

  DirectPosition2D get start => _start;
  DirectPosition2D get end => _end;


}

class XStructure {
  final SplayTreeMap<DirectPosition2D, Object> events = new SplayTreeMap();

  bool get isEmpty=> events.isEmpty;

  DirectPosition2D popNext() {
    var e = events.firstKey();
    events.remove(e);
    return e;
  }
}


class YStructure {


}

class Scene {
  List<LineSegment> _segments;
  SplayTreeMap<DirectPosition2D, List<LineSegment>> _segmentIndex;

  _addIndexFor(segment) {
    var l = _segmentIndex[segment.start];
    if (l == null) {
      l = [];
      _segmentIndex[segment.start] = l;
    }
    l.add(segment);

    l = _segmentIndex[segment.end];
    if (l == null) {
      l = [];
      _segmentIndex[segment.end] = l;
    }
    l.add(segment);
  }

  _removeIndexFor(segment){
    var l = _segmentIndex[segment.start];
    if (l == null) return;
    l.remove(segment);

    l = _segmentIndex[segment.end];
    if (l == null) return;
    l.remove(segment);
  }

  Scene() {
    _segments = [];
    _segmentIndex = new SplayTreeMap();
  }

  add(segments) {
    if (segments == null) return;
    if (segments is LineSegment) {
      segments = [segments];
    } else if (segments is List<LineSegment>) {
      // OK
    } else {
      _require(false, "expected a segment or a list of segments");
    }
    segments.forEach((s) {
      _segments.add(s);
      _addIndexFor(s);
    });
  }

  /**
   * The list of distinct positions where line segments in this
   * scene either start or end.
   */
  Iterable<DirectPosition2D> get position => _segmentIndex.keys;
}

const MAX_POS2D = const DirectPosition2D(double.INFINITY, double.INFINITY);
const MIN_POS2D = const DirectPosition2D(-double.INFINITY, -double.INFINITY);
final MAX_LINE_SEGMENT = new LineSegment(MAX_POS2D, MAX_POS2D);

class LineIntersector {
   var sweep;
   var xvalue;
   var segmentQueue;
   var xstructure;

   LineSegment dequeueSegment() {
     if (segmentQueue.isEmpty) return null;
     var key  = segmentQueue.firstKey();
     var value = segmentQueue[key];
     segmentQueue.remove(key);
     return value;
   }

  _init(segments) {
    sweep = MIN_POS2D;
    // orders line segments by the x-coordinate of their start point.
    int compareLineSegment(LineSegment s1, LineSegment s2) =>
        s1.start.x.compareTo(s2.start.x);

    // the queue of segments
    segmentQueue = new SplayTreeMap<LineSegment, dynamic>(compareLineSegment);
    segments.forEach((s) => segmentQueue.add(s, null));

    xstructure = new XStructure();
  }

  handlePassingAndEndingSegments() {
    // handle all segements passing through or ending in point sweep
    var sit = xvalue;
    if (sit == null) {
      //sit = Ystructure.locatePred()
    } else {

    }
  }

  insertStartingSegments() {
    while(true) {
      var segment = dequeueSegment();
      if (segment == null) {
        // should not happen - throw exception
      }
      if (sweep != segment.start) break;

    }
  }

  computeIntersections(Iterator<LineSegment> segments) {
    _init(segments);
    while(!xstructure.isEmpty) {
      sweep = xstructure.events.firstKey();
      xvalue = xstructure.events[sweep];
      xstructure.remove(sweep);

       handlePassingAndEndingSegments();
       insertStartingSegments();

    }
  }
}




